# -*- coding: utf-8 -*-
"""
Created on Sun Dec  5 14:04:22 2021

@author: Darshan Rathod
"""

import numpy as _np
import pandas as _pd
import datetime as _dtm
import os as _os

from .base import davis_set as _ds
from .utils import *
from . import shared as _shared


class velocity_set(_ds):
    '''
    velocity_set object which provides functionalities to work with velocity and image set of a perticular case.
    Images can only be accessed if the .set file of velocity is within the subdirectories of recording folder.
    '''

    def __init__(self, filepath, load=True, rec_path=None, load_rec=False):
        '''

        Parameters
        ----------
        filepath : string or os.path object
            path to .set file or folder
        load :bool, optional, default = True
            whether to load the file. lvreader by default does not load the buffer into the memory. If this is True then buffer is loaded into the memory.
        rec_path : string or os.path object, optional, default= None
            path to .set file of recording folder. If this is supplied then recording image can directly be loaded by calling image method.
        load_rec : BOOL, optional, default= False
            whether to load the recording set or not. MAke it True when the parent folders of .set file contains the recording set, if not then the error will arise.

        Returns
        -------
        velocity_set object

        '''
        super().__init__(filepath=filepath, load=load, rec_path=rec_path)
        if load_rec:
            self.recording_set = _ds(self.recording_foldpath)

    def __repr__(self):
        return f'velocity_set object'

    def __len__(self):
        return len(self.s)

    def __getitem__(self, i):
        return self.s[i]

    def image(self, n: int = 0, frame: int = 0):
        '''
        Returns the nth recorded image, and nth frame of the image. Generally there are 2 frames in PIV.

        Parameters
        ----------
        n : int, optional
            image number. The default is 0.
        frame : int, optional
            frame number of given image. The default is 0.

        Returns
        -------
        Image as numpy masked array.

        '''
        return self.recording_set[n][frame].as_masked_array()

    def image_masks(self, n: int = 0, frame: int = 0):
        '''
        Returns the mask of an Image.

        Parameters
        ----------
        n : int, optional
            image number. The default is 0.
        frame : int, optional
            frame number of given image. The default is 0.

        Returns
        -------
        Image mask as numpy masked array.

        '''

        return self.recording_set[n][frame].masks[0]

    def image_plot(self, n=0, frame=0, **kwargs):
        '''


        Parameters
        ----------
        n : int, optional
            image number. The default is 0.
        frame : int, optional
            frame number of given image. The default is 0.
        **kwargs : TYPE
            DESCRIPTION.

        Returns
        -------
        matplotlib.pyplot like
            plots the image

        '''
        return self.recording_set[n][frame].plot(**kwargs)

    def image_offsets(self, n=0, frame=0):
        return self.recording_set[n][frame].scales.x.offset, self.recording_set[n][frame].scales.y.offset

    def image_slopes(self, n=0, frame=0):
        return self.recording_set[n][frame].scales.x.slope, self.recording_set[n][frame].scales.y.slope

    def image_masks(self, n=0, frame=0):
        return self.recording_set[n][frame].masks[0]

    def limits(self, n=0, frame=0):
        xs, ys = self.image_slopes(n=n, frame=frame)
        xo, yo = self.image_offsets(n=n, frame=frame)
        lny, lnx = self.image().shape
        dict1 = {}
        dict1['xlim'] = [xo, xo + lnx*xs]
        dict1['ylim'] = [yo+lny*ys, yo]
        return dict1

    def image_coords(self, n=0, frame=0):
        '''
        Returns the meshgrid of X and Y coordinates of image.

        '''
        xs, ys = self.image_slopes(n=n, frame=frame)
        xo, yo = self.image_offsets(n=n, frame=frame)
        lny, lnx = self.image(n=0, frame=0).shape
        xc = _np.linspace(xo, xo+lnx*xs, lnx)
        yc = _np.linspace(yo, yo+lny*ys, lny)
        # xc,yc = rotate_coordinates_degrees(x=xc,y=yc,angle=angle)

        return _np.meshgrid(xc, yc)

    def vector_coords(self, n=0):
        '''
        Returns the meshgrid of X and Y coordinates of vectors
        '''
        xs, ys = self.vector_slopes(n=n)
        xo, yo = self.vector_offsets(n=n)
        lny, lnx = self.u(n=n).shape
        gx, gy = self.grid(n=n)
        xc = _np.linspace(xo, xo+lnx*gx*xs, lnx)
        yc = _np.linspace(yo, yo+lny*gy*ys, lny)
        xc, yc = _np.meshgrid(xc, yc)
        # xc,yc = rotate_coordinates_degrees(x=xc,y=yc,angle=angle)
        return xc, yc

    def vector_linspace(self, n=0):
        xs, ys = self.vector_slopes(n=n)
        xo, yo = self.vector_offsets(n=n)
        lny, lnx = self.u(n=n).shape
        gx, gy = self.grid(n=n)
        xc = _np.linspace(xo, xo+lnx*gx*xs, lnx)
        yc = _np.linspace(yo, yo+lny*gy*ys, lny)
        return xc, yc[::-1]

    def x(self, n=0):
        x, y = self.vector_coords(n=n)
        x = _np.ma.masked_array(x, mask=self.vector_masks(
            n=n), fill_value=0, dtype='float64')
        return x

    def y(self, n=0):
        x, y = self.vector_coords(n=n)
        y = _np.ma.masked_array(y, mask=self.vector_masks(
            n=n), fill_value=0, dtype='float64')
        return y

    def vector_offsets(self, n=0):
        return self.s[n][0].scales.x.offset, self.s[n][0].scales.y.offset

    def vector_slopes(self, n=0):
        return self.s[n][0].scales.x.slope, self.s[n][0].scales.y.slope

    def vector_masks(self, n=0):
        # return (self.s[n][0].masks[0] == 0).astype('int')
        return self.s[n][0].masks[0] == 0

    def get_keys_list(self, n=0):
        return list(self.s[n][0][0].keys())

    def get_component(self, key, n=0):
        return self.s[n][0].components[key]

    def get_arr(self, n=0, key='U0'):
        '''
        Get the array of required scalar.
        ex. key = 'U0' gives the Vx component. 
        key = 'TKE' gives the turbulent kinetic energy if it is computed in the .set file.
        '''
        # arr1 = self.s[n][0][0][key]
        # arr1 = _np.ma.masked_array(arr1,mask=self.vector_masks(n=n),fill_value=0,dtype='float64')

        comp1 = self.get_component(key=key, n=n)
        arr1 = comp1[0]
        arr1 = _np.ma.masked_array(arr1, mask=self.vector_masks(
            n=n), fill_value=0, dtype='float64')
        arr1 = comp1.scale.offset + arr1 * comp1.scale.slope
        return arr1

    def u(self, n=0):
        # return self.s[n][0].as_masked_array()['u']
        # return self.s[n][0].scales.i.offset + self.get_arr(n=n,key='U0') * self.s[n][0].scales.i.slope
        return self.get_arr(n=n, key='U0')

    def v(self, n=0):
        # return self.s[n][0].as_masked_array()['v']
        # return self.s[n][0].scales.i.offset + self.get_arr(n=n,key='V0') * self.s[n][0].scales.i.slope
        return self.get_arr(n=n, key='V0') * -1

    def TKE(self, n=0):
        # return self.s[n][0].components['TS:Turbulent kinetic energy'][0]
        return self.get_arr(n=n, key='TS:Turbulent kinetic energy')

    def plot(self, n=0):
        return self.s[n][0].plot()

    def grid(self, n=0):
        return self.s[n][0].grid.x, self.s[n][0].grid.y

    def make_contour_data(self, n=0, z='u', unit=False):
        '''
        Parameters
        ----------
        n = frame number
        z : TYPE, string
            DESCRIPTION. value for plotting. either from 'u,v,velocity_magnitude'
        unit: Type, Bool
            Description. if unit for z in output data is required. default False

        Returns
        -------
        dictionary of 'x,y,z,unit'

        '''
        data = {}
        data['x'], data['y'] = self.vector_coords(n=n)

        if (z == 'u') or (z == 'U0'):
            data['z'] = self.u(n=n)
            z = 'U0'
        elif (z == 'v') or (z == 'V0'):
            data['z'] = self.v(n=n)
            z = 'V0'
        elif z == 'velocity_magnitude':
            data['z'] = _np.sqrt(self.u(n=n)**2 + self.v(n=n)**2)
            z = 'U0'
        elif z == 'TKE':
            data['z'] = self.TKE(n=n)
        elif (z.upper() == 'WZ') or (z.upper() == 'OMEGA_Z'):
            data['z'] = self.omega_z(n=n)
        else:
            data['z'] = self.get_arr(n=n, key=z)

        if unit:
            if (z.upper() == 'WZ') or (z.upper() == 'OMEGA_Z'):
                data['unit'] = '1/s'
            else:
                data['unit'] = self.get_component(key=z, n=n).scale.unit

        return data

    def make_data(self, n=0):
        data = {}
        data['x'], data['y'] = self.vector_coords(n=n)
        data['u'], data['v'] = self.u(n=n), self.v(n=n)
        return data

    def make_image_data(self, n=0, frame=0):

        data = {}
        data['x'], data['y'] = self.image_coords(n=n, frame=frame)
        data['z'] = self.image(n=n, frame=frame)

        return data

    def make_streamline_data(self, n=0):
        data = {}

        data['u'], data['v'] = self.u(n=n), self.v(n=n)
        data['x'], data['y'] = self.vector_linspace(n=n)
        return data


def save_set(vel_set, set_name, set_foldpath, n_start=0, n_end=-1, print_info=True):
    ls1 = local_set(set_name, set_foldpath, make_folder=True)
    ls1.save_case(case=vel_set, n_start=n_start,
                  n_end=n_end, print_info=print_info)


class local_set:

    def __init__(self, set_name, set_foldpath, make_folder=False):
        '''
        class to access locally saved data. The class provides major functionalities as elocity_set, mainly make_data() function.


        Parameters
        ----------
        set_name : str
            name of the velocity_set.
        set_path : str or path type
            path of the velocity_set folder
        make_folder : BOOL, optional
            whether to make folder or not. If class is used to save the data then this attribute should be True. The default is False.

        Returns
        -------
        None.

        '''
        foldpath = _os.path.join(set_foldpath, set_name)
        self.xfp = _os.path.join(foldpath, 'x.npy')
        self.yfp = _os.path.join(foldpath, 'y.npy')
        self.Usfp = _os.path.join(foldpath, 'Us.npy')
        self.Vsfp = _os.path.join(foldpath, 'Vs.npy')
        self.maskfp = _os.path.join(foldpath, 'mask.npy')
        self.Ufp = _os.path.join(foldpath, 'Us')
        self.Vfp = _os.path.join(foldpath, 'Vs')

        if make_folder:
            make_dir(foldpath)
            make_dir(self.Ufp)
            make_dir(self.Vfp)
            # make_dir(self.SVDfp)
            self.req_loaded = False
        else:
            self.load_reqs()

    @classmethod
    def from_foldpath(cls, foldpath, *args, **kwargs):
        set_name = _os.path.basename(foldpath)
        set_foldpath = _os.path.dirname(foldpath)
        return cls(set_name, set_foldpath, *args, **kwargs)

    def __repr__(self):
        return 'local_set object'

    def __len__(self):
        return len(get_file_name(self.Ufp))

    def __getitem__(self, i):
        return self.make_data(i)

    def save_coords(self, data):
        _np.save(file=self.xfp, arr=data['x'])
        _np.save(file=self.yfp, arr=data['y'])

    def save_mask(self, data):
        _np.save(file=self.maskfp, arr=data['u'].mask)

    def load_reqs(self):
        self.x = _np.load(self.xfp)
        self.y = _np.load(self.yfp)
        self.mask = _np.load(self.maskfp)
        self.req_loaded = True

    def u(self, n=0):
        fname = str(n) + '.npy'
        return _np.ma.masked_array(_np.load(_os.path.join(self.Ufp, fname)),
                                   mask=self.mask,
                                   fill_value=0, dtype='float64')

    def v(self, n=0):
        fname = str(n) + '.npy'
        return _np.ma.masked_array(_np.load(_os.path.join(self.Vfp, fname)),
                                   mask=self.mask,
                                   fill_value=0, dtype='float64')

    def add_mask(self, data):
        return _np.ma.masked_array(data, mask=self.mask, fill_value=0, dtype='float64')

    def save_uv(self, data, n=-1):
        if n == -1:
            fname = '_1.npy'
        else:
            fname = str(n) + '.npy'
        _np.save(file=_os.path.join(self.Ufp, fname),
                 arr=data['u'].data)
        _np.save(file=_os.path.join(self.Vfp, fname),
                 arr=data['v'].data)

    def save_case(self, case, n_start=0, n_end=-1, print_info=True):
        d1 = case.make_data(n=n_start)
        self.save_coords(d1)
        self.save_mask(d1)
        if n_end == -1:
            n_end = len(case)
        if print_info:
            print('saving velocities')
        for i in range(n_start, n_end):
            if print_info:
                print(i, end='')

            d1 = case.make_data(n=i)
            self.save_uv(d1, n=i)

            if print_info:
                print('Done')
        return

    def make_data(self, n=0):
        data = {}
        if not self.req_loaded:
            self.load_reqs()
        data['x'], data['y'] = self.x, self.y
        data['u'], data['v'] = self.u(n=n), self.v(n=n)
        return data

    def make_streamline_data(self, n=0):
        data = self.make_data(n=n)
        data['x'], data['y'] = meshgrid_to_linspace(data['x'], data['y'])
        return data

    def get_multiple_u(self, n_start=0, n_end=-1):
        if n_end == -1:
            n_end = len(self)
        n = n_start
        fname = str(n) + '.npy'
        u1 = _np.load(_os.path.join(self.Ufp, fname))
        for n in range(n_start+1, n_end):
            fname = str(n) + '.npy'
            utemp = _np.load(_os.path.join(self.Ufp, fname))
            u1 = _np.dstack((u1, utemp))
        return u1

    def get_multiple_v(self, n_start=0, n_end=-1):
        if n_end == -1:
            n_end = len(self)
        n = n_start
        fname = str(n) + '.npy'
        u1 = _np.load(_os.path.join(self.Vfp, fname))
        for n in range(n_start+1, n_end):
            fname = str(n) + '.npy'
            utemp = _np.load(_os.path.join(self.Vfp, fname))
            u1 = _np.dstack((u1, utemp))
        return u1

    def save_UVs(self, n_start=0, n_end=-1):
        u1 = self.get_multiple_u(n_start=n_start, n_end=n_end)
        _np.save(file=self.Usfp,
                 arr=u1)

        u1 = self.get_multiple_v(n_start=n_start, n_end=n_end)
        _np.save(file=self.Vsfp,
                 arr=u1)
        

    @property
    def Us(self):
        return _np.load(file=self.Usfp)

    @property
    def Vs(self):
        return _np.load(file=self.Vsfp)



# -*- coding: utf-8 -*-
"""
Created on Fri Jan  7 21:04:55 2022

@author: darsh
"""

import matplotlib as _mpl
import matplotlib.pyplot as _plt
import numpy as _np
import pandas as _pd


def remove_ax_lines(ax):
    for spine in ['top', 'right','bottom','left']:
        ax.spines[spine].set_visible(False)
    return ax

# def plot_quiver(vel_set=None,n=0,data=None,ax=None,fracx=6,fracy=None,scale=0.8,width=0.1,headwidth=12,headlength=15,minshaft=2,minlength=0.1,units='xy',scale_units='xy'):
def plot_quiver(vel_set=None,n=0,data=None,ax=None,fracx=3,fracy=None,scale=50,width=0.0015,headwidth=4,normalize=True,**kwargs):
    '''
    plots quiver (vectors) either from vel_set or data, whichever is defined. If vel_set is defined then n (image number), 'z'(saclar to plot) is to be
    defined. If both vel_set and data are defined then data is given the priority above vel_set.

    Parameters
    ----------
    vel_set : velocity_set, optional
        either provide the velocity set or the data for plotting. The default is None.
    n : int, optional
        image/frame number. This is used when vel_set is supplied to the function. If data is directly given then this option is not useful. The default is 0.
    data : dict, optional
        dictionary-like object containing 'x','y',and 'z' keys and numpy_masked array like values. The 'z' values are contour filled in 'x' and 'y' coordinates.
        This is optional if the vel_set is defined. The default is None.
    ax : matplotlib.pyplot.axes like, optional
        axes on which to plot. The default is None. If not specified then plt.gca() is used for plotting.
    fracx : int, optional
        sub-sample over the x-axis to reduce the clutter. If fracx is 4 then for every 4 data one will be taken for plotting. The default is 3.
    fracy : str or int, optional
        Same as fracx but in y-direction. The default is None. If None then fracy = fracx
    normalize : bool, optional
        whether to normalize the vector for plotting. if true, vectors are normalized with its magnitude. make this true if vectors vary too much in magnitude over the plot.
    **kwargs : plt.quiver keyword arguments
        gives the freedom to modify the plot according to the needs.

    Returns
    -------
    ax : matplotlib.pyplot.axes like
        returns the axes with plot

    '''
    if ax is None:
        ax = _plt.gca()
    
    if data is None:
        data = vel_set.make_data(n=n)
    
    if fracy is None:
        fracy = fracx
        
    x= data['x']
    y = data['y']
    u = data['u']
    v = data['v']
    idx = []
    for i in range(0,x.shape[0],fracy):
        idx.append(i)
    x1 = x[idx]
    y1 = y[idx]
    u1 = u[idx]
    v1 = v[idx]
    
    idx = []
    for i in range(0,x1.shape[1],fracx):
        idx.append(i)
    x1 = x1[:,idx]
    y1 = y1[:,idx]
    u1 = u1[:,idx]
    v1 = v1[:,idx]
    
    data = {'x': x1, 'y':y1,
            'u': u1, 'v':v1}
    
    if normalize:
        data['z'] = (data['u']**2+data['v']**2)**0.5
        data['u'] = data['u']/data['z']
        data['v'] = data['v']/data['z']
    
    ax.quiver(data['x'],data['y'],data['u'],data['v'],scale=scale,width=width,headwidth=headwidth,**kwargs)
    return ax
    
def plot_colorbar(ax=None,cax=None,vmax='max',vmin='min',colormap=None,
                  ctitle=None,font_size=None,cticks=11,roundto=1,clabel=None,rotation=270,labelpad=10,
                  titlepad=10, ticklabels:list=None):
    if vmax == 'max':
        vmax = 1
    if vmin == 'min':
        vmin = 0
    
    cmap = _plt.get_cmap(colormap,256)
    norm = _mpl.colors.Normalize(vmin=vmin,vmax=vmax)
    sm = _plt.cm.ScalarMappable(cmap=cmap,norm=norm)
    
    ticks1 =  _np.linspace(vmin,vmax,cticks,endpoint=True).round(roundto)
    
    if cax is None:
        if ax is None:
            ax = _plt.gca()
        cbar = _plt.colorbar(sm,ax=ax,ticks=ticks1,orientation='vertical')
    else:
        cbar = _plt.colorbar(sm,cax=cax,ticks=ticks1,orientation='vertical')
    
    # cbar.set_ticklabels(ticks1)
    cbar.set_ticks(ticks1)
    if ctitle is not None:
        cbar.ax.set_title(ctitle,fontsize=font_size,pad=titlepad)
    if clabel is not None:
        cbar.set_label(clabel,rotation=rotation,labelpad=labelpad)
    if font_size is not None:
        cbar.ax.tick_params(labelsize=font_size)
    if ticklabels is not None:
        cbar.ax.set_yticklabels(ticklabels)
    return ax


def plot_contourf(vel_set=None,n=0,data=None,z='u',ax=None,vmax='max',vmin='min',
                  add_colorbar=True,colormap=None,ctitle=None,font_size=None,cticks=10,levels=200,alpha=1,
                  roundto=1,clabel=None,rotation=270,labelpad=10,qtile=0.01,equalize_at='min'):
    '''plot contourf for the velocity_set
    plots filled contour of a scalar either from vel_set or data, whichever is defined. If vel_set is defined then n (image number), 'z'(saclar to plot) is to be
    defined. If both vel_set and data are defined then data is given the priority above vel_set.

    Parameters
    ----------
    vel_set : velocity_set, optional
        either provide the velocity set or the data for plotting. The default is None.
    n : int, optional
        image/frame number. This is used when vel_set is supplied to the function. If data is directly given then this option is not useful. The default is 0.
    data : dict, optional
        dictionary-like object containing 'x','y',and 'z' keys and numpy_masked array like values. The 'z' values are contour filled in 'x' and 'y' coordinates.
        This is optional if the vel_set is defined. The default is None.
    z : str, optional
        scalar to plot, it can be from 'u,v,velocity_magnitude, omega_z, Wz, TKE'. The default is 'u'. This is used only when vel_set is defined.
    ax : matplotlib.pyplot.axes like, optional
        axes on which to plot. The default is None. If not specified then plt.gca() is used for plotting.
    vmax : float, optional
        maximum value in colorbar. The default is 'max'.
    vmin : float, optional
        minimum values in colorbar. The default is 'min'.
    add_colorbar : TYPE, optional
        DESCRIPTION. The default is True.
    colormap : TYPE, optional
        DESCRIPTION. The default is 'jet'.
    ctitle : TYPE, optional
        DESCRIPTION. The default is ''.
    font_size : TYPE, optional
        DESCRIPTION. The default is None.
    cticks : TYPE, optional
        DESCRIPTION. The default is 10.
    levels : TYPE, optional
        DESCRIPTION. The default is 200.
    alpha : TYPE, optional
        DESCRIPTION. The default is 1.
    roundto : TYPE, optional
        DESCRIPTION. The default is 2.
    qtile : float, optional
        DESCRIPTION. The default is 0.01. This is the quantile value to calculate vmax and vmin from z data.
    equalize_at : str, optional
        DESCRIPTION. The default is 'min'. The colorbar is equalized such that zero remains at the center. if 'min' then min(abs(vamx),abs(vmin)) is taken and if
        'max' then max is taken.

    Returns
    -------
    ax : matplotlib.pyplot.axes like
        returns the axes with plot

    '''
    
    if ax is None:
        ax = _plt.gca()
    if data is None:
        data = vel_set.make_contour_data(n=n,z=z)
    if (vmax == 'max') and (vmin=='min'):
        if qtile == 0:
            vmax = data['z'].data.max()
            vmin = data['z'].data.min()
        else:
            vmax = _np.quantile(data['z'].data,1-qtile).round(roundto)
            vmin = _np.quantile(data['z'].data,qtile).round(roundto)
        
        if (vmin==0) or (vmax==0):
            pass
        elif ((vmax /abs(vmax)) == (vmin/abs(vmin))):
            pass
        else:
            if equalize_at == 'min':
                v1 = min(abs(vmax),abs(vmin))
            if equalize_at == 'max':
                v1 = max(abs(vmax),abs(vmin))
            vmax = v1 * vmax / abs(vmax)
            vmin = v1 * vmin / abs(vmin)
    if vmax == 'max':
        vmax = _np.quantile(data['z'].data,0.99).round(roundto)
    if vmin == 'min':
        vmin = _np.quantile(data['z'].data,0.01).round(roundto)
    
    cmap = _plt.get_cmap(colormap,256)
    norm = _mpl.colors.Normalize(vmin=vmin,vmax=vmax)
    
    cp = ax.contourf(data['x'],data['y'],data['z'],
                     levels=levels,cmap=cmap,norm=norm,corner_mask=True,alpha=alpha)
    
    if add_colorbar:
        plot_colorbar(ax=ax,vmax=vmax,vmin=vmin,colormap=colormap,ctitle=ctitle,
                     font_size=font_size,cticks=cticks,roundto=roundto,clabel=clabel,rotation=rotation,labelpad=labelpad)
    
    return ax


def plot_image(vel_set=None,n=0,frame=0,data=None,ax=None,vmin=0,vmax=3000,levels=20,colormap='viridis',alpha=1):
    
    if ax is None:
        ax = _plt.gca()
    
    if data is None:
        data = vel_set.make_image_data(n=n,frame=frame)
    
    cmap = _plt.get_cmap(colormap,256)
    norm = _mpl.colors.Normalize(vmin=vmin,vmax=vmax)
    
    cp = ax.contourf(data['x'],data['y'],data['z'],levels=levels,corner_mask=True,
                     cmap=cmap,norm=norm,alpha=alpha)
    return ax


def plot_streamlines(vel_set=None,n=0,data=None,ax=None,density=(5,5),linewidth=1,color='white',**kwargs):
    
    if ax is None:
        ax = _plt.gca()
    
    if data is None:
        data = vel_set.make_streamline_data(n=n)
    
    ax.streamplot(data['x'],data['y'],data['u'][::-1],data['v'][::-1],
                  density=density,linewidth=linewidth,color=color,**kwargs)
    
    return ax






# -*- coding: utf-8 -*-
"""
Created on Sun Dec  5 14:04:37 2021

@author: Darshan Rathod
"""


import matplotlib as _mpl
import matplotlib.pyplot as _plt
import numpy as _np
import pandas as _pd
import os as _os

from . import shared as _shared


def make_dir(path1):
    try:
        _os.mkdir(path1)
    except FileExistsError as e:
        pass


def get_dir_path(fname):
    return [_os.path.join(fname,x) for x in _os.listdir(fname) if _os.path.isdir(_os.path.join(fname,x))]

def get_file_path(fname):
    return [_os.path.join(fname,x) for x in _os.listdir(fname) if _os.path.isfile(_os.path.join(fname,x))]

def get_dir_name(fname):
    return [x for x in _os.listdir(fname) if _os.path.isdir(_os.path.join(fname,x))]

def get_file_name(fname):
    return [x for x in _os.listdir(fname) if _os.path.isfile(_os.path.join(fname,x))]



def move_origin(d1,xnew=0, ynew=0):
    d1['x'] = d1['x'] - xnew
    d1['y'] = d1['y'] - ynew
    return d1

def rotate_coordinates_degrees(x,y,angle=0):
    xy1 = _np.array([x.flatten(),y.flatten()])
    cos1 = _np.cos(angle*_np.pi/180)
    sin1 = _np.sin(angle*_np.pi/180)
    Rv = _np.array([[cos1,-sin1],[sin1,cos1]])
    xy1 = _np.matmul(Rv,xy1)
    x = xy1[0,:].reshape(x.shape)
    y = xy1[1,:].reshape(y.shape)
    return x,y

def rotate_bases(d1,angle=0):
    '''Rotates the coordinate bases

    Parameters
    ----------
    d1 : TYPE
        DESCRIPTION.
    angle : TYPE, optional
        DESCRIPTION. The default is 0.

    Returns
    -------
    d1 : TYPE
        DESCRIPTION.

    '''
    
    mask1 = d1['u'].mask
    uv = _np.array([d1['u'].data.flatten(),d1['v'].data.flatten()])
    theta1 = angle * _np.pi/180
    A = _np.array([[_np.cos(theta1),_np.sin(theta1)],[-_np.sin(theta1),_np.cos(theta1)]]).T
    Ainv = _np.linalg.inv(A)
    uv1 = _np.matmul(Ainv,uv)
    d1['u'] = uv1[0,:].reshape(d1['u'].shape)
    d1['u'] = _np.ma.masked_array(d1['u'],mask=mask1,fill_value=0,dtype='float64')
    d1['v'] = uv1[1,:].reshape(d1['v'].shape)
    d1['v'] = _np.ma.masked_array(d1['v'],mask=mask1,fill_value=0,dtype='float64')
    return d1

def get_data_at_point(data,px=0,py=0,dx=None,dy=None):
    '''
    calculates the data['z'] at (px,py) point in the domain. Data is interpolated using distance weighted average from its neighbouring points.

    Parameters
    ----------
    data : dict
        s1.make_data() type of data. It has x,y,z values. x,y are coordinates meshgrid. and z is the scalar of interest in the domain.
    px : float, optional
        X coordinate. The default is 0.
    py : float, optional
        Y coordinate. The default is 0.
    dx : float, optional
        delta x in the data['x']. This value is utilized to filter out the neighbours. The code looks at px +/- dx coordinates for all the points.
        then from obtained coordinates, it calculates the weighted average. The default is None. If None then code calculated dx from itself.
    dy : float, optional
        delta y in the data['y']. The default is None.

    Returns
    -------
    pz : float
        weighted average of data['z'].

    '''
    if dx is None:
        dx = _np.diff(_np.unique(data['x']))[0]
    if dy is None:
        dy = _np.diff(_np.unique(data['y']))[0]
        
    f1 = data['x'] >= (px - dx)
    f1 = f1 & (data['x'] <= (px + dx))
    f1 = f1 & (data['y'] >= (py - dy))
    f1 = f1 & (data['y'] <= (py + dy))
    
    px1 = data['x'][f1]
    py1 = data['y'][f1]
    pz1 = data['z'].data[f1]
    
    ws = ((px1 - px)**2 + (py1 - py)**2)**0.5
    pz = (ws*pz1).sum()/ws.sum()
    return pz

def make_line(x1:float,y1:float,x2:float,y2:float,n_points:int=100):
    '''
    generates coordinates (=n_points) over the line passing from (x1,y1) and (x2,y2)

    Parameters
    ----------
    x1 : float
        x - coordinate
    y1 : float
        y - coordinate
    x2 : float
        x - coordinate
    y2 : float
        y - coordinate
    n_points : int, optional
        number of points between (x1,y1) and (x2,y2). The default is 100.

    Returns
    -------
    x : numpy_array like
        x-coordinates
    y : numpy_array like
        y-coordinates

    '''
    # if the ilne is verticle then "division by zero" error will come
    if x2 == x1:
        x = _np.zeros(n_points) + x2
        y = _np.linspace(y1,y2,n_points)
    else:
        m = (y2-y1) / (x2-x1)
        c = y2 - m*x2
        x = _np.linspace(x1,x2,n_points)
        y = m*x + c
    return x,y


def get_data_at_line(data,x1,y1,x2,y2,n_points=100,dx=None,dy=None):
    '''
    generates coordinates (=n_points) over the line passing from (x1,y1) and (x2,y2) and calculates the interpolated data['z'] over the line.
    
    The code utilizes the function make_line() to get line coordinates, and get_data_at_point() function to get data at individual point over the line.
    Hence the detail explanation of all the arguments of this function can be obtained from the above two functions.

    Parameters
    ----------
    x1 : float
        x - coordinate
    y1 : float
        y - coordinate
    x2 : float
        x - coordinate
    y2 : float
        y - coordinate
    n_points : int, optional
        number of points between (x1,y1) and (x2,y2). The default is 100.
    dx : float, optional
        delta x in the data['x']. This value is utilized to filter out the neighbours. The code looks at px +/- dx coordinates for all the points.
        then from obtained coordinates, it calculates the weighted average. The default is None. If None then code calculated dx from itself.
    dy : float, optional
        delta y in the data['y']. The default is None.    

    Returns
    -------
    x : numpy_array like
        x-coordinates
    y : numpy_array like
        y-coordinates
    z : numpy_array like
        data['z'] values over x and y

    '''
    x,y = make_line(x1,y1,x2,y2,n_points=n_points)
    z = _np.zeros(x.shape)
    i = 0
    for xp,yp in zip(x,y):
        z[i] = get_data_at_point(data,px=xp,py=yp,dx=dx,dy=dy)
        i += 1
    return x,y,z


def get_omega_z(data):
    dx = _np.diff(_np.unique(data['x']))[0]
    dy = _np.diff(_np.unique(data['y']))[0]
    wz = (_np.gradient(-data['v'],dx,axis=1) - _np.gradient(data['u'],dy,axis=0))*1000
    data['z'] = -wz
    return data

def get_mod_V(data):
    data['z'] = (data['u']**2 + data['v']**2)**0.5
    return data


def imshow(img,*args,**kwargs):
    _plt.imshow(img,*args,**kwargs)
    _plt.xticks([])
    _plt.yticks([])

def meshgrid_to_linspace(x,y):
    '''
    converts x and y meshgrid to numpy.linspace type array

    Parameters
    ----------
    x : 2-D numpy array
    y : 2-D numpy array

    Returns
    -------
    x1 : 1-D numpy array
    y1 : 1-D numpy array

    '''
    x1 = _np.linspace(x.min(),x.max(),x.shape[1])
    y1 = _np.linspace(y.min(),y.max(),y.shape[0])
    return x1,y1

def get_streamline_data(data):
    data['x'],data['y'] = meshgrid_to_linspace(data['x'],data['y'])
    return data


def sample_at_point(
    arr: _np.ndarray,
    d1_avg: dict,
    px: float,
    py: float,
    bbox_size: tuple[int, int] = (0, 0),
) -> dict:
    """Samples the 3-D array at (px,py) point.

    The code samples the input data in arr. The information about the location of the point is obtained from d1_avg 'x' and 'y' key.

    Parameters
    ----------
    arr : np.ndarray
        array to be sampled
    d1_avg : dict
        dictionary containing 'x','y' and 'z' key.
    px : float
        x-coordinate of the sample point in mm.
    py : float
        y-coordinate of the sample point in mm.
    bbox_size : tuple[int, int], optional
        size of the box around (px,py) for sampling in mm, by default (0, 0)

    Returns
    -------
    dict
        dictionary containing 'x','y' and 'z' keys. 'x' and 'y' contains the coordinates of the sampled points. 'z' contains the sampled values.
    """

    x0, y0 = px, py
    bbox_size_mm = 1, 1  # in mm

    f1 = (d1_avg["x"] < (x0 + bbox_size_mm[0])) & (d1_avg["x"] > (x0 - bbox_size_mm[0]))
    f1 = (
        f1
        & (d1_avg["y"] < (y0 + bbox_size_mm[1]))
        & (d1_avg["y"] > (y0 - bbox_size_mm[1]))
    )
    px, py = _np.where(f1)
    # print(px, py)
    px, py = px[0], py[0]

    if bbox_size == (0, 0):
        arr2 = arr[px, py, :]
        x = d1_avg["x"][px, py]
        y = d1_avg["y"][px, py]
    else:
        arr2 = arr[
            px - bbox_size[0] : px + bbox_size[0],
            py - bbox_size[1] : py + bbox_size[1],
            :,
        ]
        x = d1_avg["x"][
            px - bbox_size[0] : px + bbox_size[0],
            py - bbox_size[1] : py + bbox_size[1],
        ]
        y = d1_avg["y"][
            px - bbox_size[0] : px + bbox_size[0],
            py - bbox_size[1] : py + bbox_size[1],
        ]

    return {"x": x, "y": y, "z": _np.array(arr2)}















# -*- coding: utf-8 -*-
"""
Created on Sat Dec  4 13:18:33 2021

@author: Darshan Rathod

This module defines davis_set object which gives basic functionalities to handle davis set files. Class calib_file extracts the calibration information of the davis set object.

"""

from lvreader import read_set as _rs
import os as _os
import pandas as _pd
import numpy as _np
import datetime as _dtm
import concurrent.futures as _conf
import xml.etree.ElementTree as _ET

from .utils import *


def analyze_set(foldpath,load=True,rec_path=None):
    s1 = davis_set(foldpath,load=load,rec_path=rec_path)
    ans1 = {'recording':False,'image':False,'vector':False,'Avg_vector':False}
    ans1[s1.data_type] = True
    ans1['recording_foldpath'] = s1.recording_foldpath
    ans1['foldpath'] = s1.foldpath
    ans1['parent_fold_len'] = s1.parent_fold_len
    return _pd.DataFrame(ans1,index=[0])




class davis_set:
    '''
    Adds functionality to handle davis set file
    '''
    
    def __init__(self,filepath: str,load: bool = True, rec_path: str = None):
        '''

        Parameters
        ----------
        filepath : string or os.path object
            path to .set file or folder
        load :bool, optional, default = True
            whether to load the file. lvreader by default does not load the buffer into the memory. If this is True then buffer is loaded into the memory.
        rec_path : string or os.path object, optional, default= None
            path to .set file of recording folder. If this is supplied then recording image can directly be loaded by calling image method.

        Returns
        -------
        davis_set object

        '''
        self.filepath = filepath
        self.s = _rs(self.filepath)
        if load:
            self.buffer = self.s[0]
        self.rec_path = rec_path
    
    @property
    def is_closed(self):
        return self.s.closed
    
    def close(self):
        return self.s.close()
    
    def __repr__(self):
        return f'davis_set object'
    
    def __len__(self):
        return len(self.s)
    
    def __getitem__(self,i):
        return self.s[i]
    
    def __type__(self):
        return 'davis_set'
    
    @property
    def project_foldpath(self):
        return _os.path.dirname(self.recording_foldpath)
    
    @property
    def project_foldname(self):
        return _os.path.basename(self.project_foldpath)
    
    
    @property
    def recording_foldpath(self):
        '''
        Returns
        -------
        path of the recording folder
        '''
        if self.rec_path != None:
            return self.rec_path
        if self.data_type == 'recording':
            return self.foldpath
        else:
            s1 = davis_set(_os.path.dirname(self.foldpath))
            return s1.recording_foldpath
    
    @property
    def recording_foldname(self):
        return _os.path.basename(self.recording_foldpath)
    
    
    @property
    def name(self):
        return _os.path.basename(self.foldpath)
    
    @property
    def data_type(self):
        str1 = self.s.type_id
        if str1 == 'SET_TYPE_ID_RECORDING':
            return 'recording'
        elif str1 == 'SET_TYPE_ID_IMAGE':
            return 'image'
        elif str1 == 'SET_TYPE_ID_VECTOR':
            if 'Avg' in self.name:
                return 'Avg_vector'
            else:
                return 'vector'
    
    @property
    def attributes(self):
        return self.buffer.attributes
    
    
    @property
    def foldpath(self):
        return self.buffer.attributes['LoadSet'] 
    
    @property
    def dt(self):
        '''
        Returns
        -------
        dt in micro-seconds

        '''
        return self.buffer.attributes['DevDataTrace2'][0,0]
    
    @property
    def cam1_exposure(self):
        '''
        Returns
        -------
        Exposure time in micro-seconds
        '''
        return self.buffer.attributes['DevDataTrace0'][0,0]
    
    @property
    def ref_time(self):
        '''
        Returns
        -------
        reference time in mili-seconds
        '''
        return self.buffer.attributes['DevDataTrace1'][0,0]
    
    @property
    def laser_powers(self):
        '''
        Returns
        -------
        laser power in % for laser A and B respectively
        '''
        return self.buffer.attributes['DevDataTrace3'][0,0] , self.buffer.attributes['DevDataTrace4'][0,0]
    
    @property
    def trigger_rate(self):
        '''
        Returns
        -------
        triggering frequency in Hz
        '''
        return self.buffer.attributes['DevDataTrace5'][0,0]
    
    @property
    def recording_rate(self):
        '''
        Returns
        -------
        recording frequency in Hz
        '''
        return self.buffer.attributes['DevDataTrace6'][0,0]
    
    @property
    def time0(self):
        '''
        Returns
        -------
        starting time as datetime.datetime object
        '''
        
        dt1 = [int(x) for x in self.buffer.attributes['_Date'].split('.')]
        lt1 = self.buffer.attributes['_Time'].split('.')
        us1 = int(lt1[1])*1000
        t1 = [int(x) for x in lt1[0].split(':')]
        return _dtm.datetime(dt1[2]+2000,dt1[1],dt1[0],t1[0],t1[1],t1[2],us1)
    
    @property
    def parent_fold_len(self):
        if self.data_type != 'recording':
            fp = _os.path.dirname(self.foldpath)
            s1 = davis_set(fp)
            return len(s1)
        else:
            return len(self.s)

    
    def get_analysis_list(self):
        ls1 = [x[0] for x in _os.walk(self.foldpath)]
        if self.data_type == 'recording':
            rec_path = self.foldpath
        else:
            rec_path = self.rec_path
        ans1 = _pd.DataFrame()
        for f in ls1:
            ans1 = ans1.append(analyze_set(f,rec_path=rec_path),ignore_index=True)
        self.analysis_list = ans1
        return ans1
    
    
    @property
    def calibration_foldpath(self):
        return _os.path.join(_os.path.dirname(self.recording_foldpath),'Properties','Calibration')
    
    def load_calibration(self):
        fp1 = _os.path.join(self.calibration_foldpath,'Calibration.xml')
        self.calibration = calib_file(fp1)
        return self.calibration
    
    
    def get_loading_func(self, comp:str):
        if comp == 'u':
            loading_func = self.u
        if comp == 'v':
            loading_func = self.v
        if comp == 'w':
            loading_func = self.w
        if comp == 'img':
            loading_func = self.image
        return loading_func
    
    def print_num(self,num):
        if (num%50==0):
            print(num)
        else:
            print(num,end=',')
        return
    
    def get_multiple_data(self,comp:str, n_start:int=0, n_end:int=-1,print_info:bool=False)->_np.array:
        if n_end == -1:
            n_end = len(self)
        loading_func = self.get_loading_func(comp=comp)
        u1 = loading_func(n=n_start)
        if print_info:
            self.print_num(n_start)
        for n in range(n_start + 1, n_end):
            if print_info:
                self.print_num(n)
            utemp = loading_func(n=n)
            u1 = _np.dstack((u1, utemp))
        return u1.data

    def save_as_single_array(self,comp:str,print_info:bool=False,filepath:str=None)->None:
        fname = comp.upper() + 's.npy'
        if filepath is None:
            fpath = _os.path.join(self.filepath,fname)
        else:
            fpath = _os.path.join(filepath,fname)
        
        n_start = 0
        n_end = len(self)

        loading_func = self.get_loading_func(comp=comp)
        u1 = loading_func(n=0)
        sh = u1.shape
        if comp == 'img':
            dtype1 = 'uint16'
        else:
            dtype1 = u1.dtype
        data = _np.memmap(filename=fpath, dtype=dtype1, mode='w+', shape=(sh[0],sh[1],n_end))
        for n in range(n_start,n_end):
            data[:,:,n] = loading_func(n=n)
            if print_info:
                self.print_num(n)
        data.flush()
        return
    
    def load_as_single_array(self,comp:str,filepath:str=None, mode:str='c')->_np.array:
        fname = comp.upper() + 's.npy'
        if filepath is None:
            fpath = _os.path.join(self.filepath,fname)
        else:
            fpath = _os.path.join(filepath,fname)
        loading_func = self.get_loading_func(comp=comp)
        u1 = loading_func(n=0)
        sh = u1.shape
        if comp == 'img':
            dtype1 = 'uint16'
        else:
            dtype1 = u1.dtype
        data = _np.memmap(filename=fpath, dtype=dtype1, mode=mode)
        n1 = int(data.shape[0] / sh[0] / sh[1])
        return data.reshape(sh[0],sh[1],n1)
    
    
    
    



class calib_file:
    
    def __init__(self,filepath):
        self.filepath = filepath
        self.tree = _ET.parse(filepath)
        self.root = self.tree.getroot()
        
    @property
    def foldpath(self):
        return _os.path.dirname(self.filepath)
    
    def get_image(self,n_camera=1):
        fp1 = _os.path.join(self.foldpath,'camera'+str(n_camera))
        s1 = davis_set(fp1)
        return s1[0][0].as_masked_array()
        
    
    def get_view(self,n_view=1):
        return self.root[n_view-1][0]
    
    def get_camera(self,n_camera=1,n_view=1):
        return self.get_view(n_view=n_view)[n_camera-1]
    
    
    def get_scales(self,n_camera=1,n_view=1):
        d1 = {}
        for i in self.get_camera(n_camera=n_camera,n_view=n_view)[0][4]:
            d1[i.tag] = i.attrib
        self.scales = _pd.DataFrame(d1)
        return self.scales
    
    def get_external_camera_parameters(self,n_camera=1,n_view=1):
        d1 = {}
        for i in self.get_camera(n_camera=n_camera,n_view=n_view)[0][3]:
            d1[i.tag] = i.attrib
        self.external_camera_parameters = d1
        return self.external_camera_parameters
    
    def get_internal_camera_parameters(self,n_camera=1,n_view=1):
        d1 = {}
        for i in self.get_camera(n_camera=n_camera,n_view=n_view)[0][2]:
            d1[i.tag] = i.attrib
        self.internal_camera_parameters = d1
        return self.internal_camera_parameters
    
    def get_common_parameters(self,n_camera=1,n_view=1):
        d1 = {}
        for i in self.get_camera(n_camera=n_camera,n_view=n_view)[0][0]:
            d1[i.tag] = i.attrib
        self.common_parameters = d1
        return self.common_parameters
    
    def get_rotation_angles(self,n_camera=1,n_view=1):
        return _pd.DataFrame(self.get_external_camera_parameters(n_camera=n_camera,n_view=n_view)['RotationAngles'],index=['Radians'],dtype='float')
    
    def get_translation_mm(self,n_camera=1,n_view=1):
        return _pd.DataFrame(self.get_external_camera_parameters(n_camera=n_camera,n_view=n_view)['TranslationMm'],index=[0],dtype='float')
        
    def get_focal_length(self,n_camera=1,n_view=1):
        return float(self.get_internal_camera_parameters(n_camera=n_camera,n_view=n_view)['FocalLengthMm']['Value'])
        
    def get_RMS_fit(self,n_camera=1,n_view=1):
        return float(self.get_common_parameters(n_camera=n_camera,n_view=n_view)['FitError']['RMS'])
    
    def get_pixcel_per_mm(self,n_camera=1,n_view=1):
        return float(self.get_common_parameters(n_camera=n_camera,n_view=n_view)['PixelPerMmFactor']['Value'])
    
    def get_calibration_plate_image(self,n_camera=1):
        fp1 = _os.path.dirname(self.filepath)
        fp1 = _os.path.join(fp1,'camera'+str(n_camera))
        self.calib_plate = _rs(fp1)[0][0].as_masked_array()
        return self.calib_plate
    
    
    @property
    def version(self):
        return self.root.attrib['Version']
    
    
class davis_project:
    
    def __init__(self,foldpath):
        self.foldpath = foldpath
    
    def __repr__(self):
        return f'davis_project object'
    
    def __len__(self):
        return len(self.all_sets)
    
    @property
    def name(self):
        return _os.path.basename(self.foldpath)
    
    @property
    def all_sets(self):
        ls1 = []
        for x in get_file_name(self.foldpath):
            if '.set' in x:
                if 'Properties' in x:
                    pass
                else:
                    ls1.append(x)
        # ls1.pop(ls1.index('Properties.set'))
        return ls1
    
    @property
    def all_sets_filepath(self):
        ls1 = []
        for x in self.all_sets:
            ls1.append(_os.path.join(self.foldpath,x))        
        return ls1
    
    @property
    def all_sets_foldpath(self):
        ls1 = []
        for x in get_dir_name(self.foldpath):
            if 'Properties' in x:
                pass
            else:
                ls1.append(_os.path.join(self.foldpath,x))        
        return ls1
    
    
    def get_analysis_list(self):
        def func1(fp):
            s1 = davis_set(fp)
            return s1.get_analysis_list()
        
        if len(self.all_sets) == 0:
            return
        
        ls1 = self.all_sets_foldpath
        ans1 = _pd.DataFrame()
        with _conf.ThreadPoolExecutor(max_workers=len(ls1)) as exe:
            for f in ls1:
                ans1 = ans1.append(exe.submit(func1,f).result(),ignore_index=True)
        self.analysis_list = ans1
        return ans1













    


# -*- coding: utf-8 -*-
"""
Created on Sun Dec  5 14:04:00 2021

@author: Darshan Rathod

This package has functionalities to manage, plot and handle davis ouput files. The major functionalities includes
- accessing *.im7* (image) files
- accessing *.vc7* (vector) files
- plotting vectors, streamlines and quiver plots
- calculation of $\omega$ from velocity vectors

The basic object are defined in *base.py* file. Other files are built upon the classes defined in *base.py* file.
All common functions are defined in *utils.py* file. Ther might be some functions in other filed which are specific to that module.

"""

from .base import *
from .PIV_2D import *
from .utils import *
from .plotting import *


from .version import version


from setuptools import setup, find_packages
import sys
import os 

module_name = 'ddr_davis_data'
gihub_url = r'https://github.com/ddrathod121294/' + module_name

p1 = os.path.join(r'src', module_name)
sys.path.insert(0, os.path.join(os.getcwd(),p1))

from version import version 

with open('README.md') as readme_file:
    readme = readme_file.read()


requirements = ["numpy",
                "pandas",
                "matplotlib",
                "openpyxl",
                ]

setup_requirements = []

test_requirements = []

setup(
    name=module_name,
    version=version,
    packages=find_packages(where="src"),  # Required
    url=gihub_url,
    description="Package to handle davis data files", # Optional
    keywords='ddr,davis,davis_data,davis_data_manager,' + module_name, # Optional

    python_requires=">=3.10, <4",
    author="Darshan Rathod", # Optional
    author_email='darshan.rathod1994@gmail.com', # Optional
    classifiers=[ # Optional
        'Development Status :: 2 - Pre-Alpha',
        
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        
        'Natural Language :: English',

        "License :: OSI Approved :: MIT License",
        
        'Topic :: Education',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Physics',
        'Programming Language :: Python :: 3.10'
    ],
    package_dir={"": "src"},  # Optional
    install_requires=requirements, # Optional 
    long_description=readme, # Optional
    long_description_content_type="text/markdown", # Optional 

    # setup_requires=setup_requirements,
    # test_suite='tests',
    # tests_require=test_requirements,
    # zip_safe=False,
    # include_package_data=True,
)


